/*
 * Author:	William Chia-Wei Cheng (william@cs.ucla.edu)
 *
 * Copyright (C) 1990, 1991, 1992, William Cheng.
 * 
 * Permission limited to the use, copy, modify, and distribute this software
 * and its documentation for any purpose is hereby granted by the Author without
 * fee, provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the Author not be used
 * in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The Author makes no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.  All other
 * rights are reserved by the Author.
 *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
 * USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef lint
static char RCSid[] =
      "@(#)$Header: /amnt/kona/tangram/u/william/X11/TGIF2/RCS/xbitmap.c,v 2.115 1992/10/21 05:23:23 william Exp $";
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "const.h"
#include "types.h"

#include "auxtext.e"
#include "choice.e"
#include "cmd.e"
#include "color.e"
#include "cursor.e"
#include "dialog.e"
#include "dup.e"
#include "eps.e"
#include "file.e"
#include "font.e"
#include "grid.e"
#include "mark.e"
#include "menu.e"
#include "msg.e"
#include "names.e"
#include "obj.e"
#include "pattern.e"
#include "raster.e"
#include "select.e"
#include "setup.e"
#include "xpixmap.e"

extern int	rand();

GC	xbmGC = NULL;

int	askForXBmSpec = FALSE;
int	stripEPSComments = TRUE;

Pixmap	dummyBitmap;

static char	hexValue[] = "0123456789abcdef";

static int	importXBmRV = FALSE;
static int	numColorsToDump = 0;
static int	charsPerPixel = 1;
static int	* pixelValue = NULL;
static int	* colorIndexToDumpIndex = NULL;
static char	* colorChar = NULL;
static char	* * colorStr = NULL;

static int	xpmOutputVersion = 1;
static int	xpmInXGrabSCFormat = FALSE;

static int	halfToneBitmap = FALSE;
static int	thresholdBitmap = FALSE;
static float	bitmapThreshold = 0.5;
static char	bitmapThresholdStr[MAXSTRING+1];

typedef struct LineRec {
   char			* s;
   struct LineRec	* next, * prev;
} * LineRecPtr;

static struct LineRec	* topLine=NULL, * botLine=NULL;
static int		numLines = 0;

static
void ExpandTmpStorage ()
{
   pixelValue = (int *) realloc (pixelValue, (maxColors+2)*sizeof(int));
   colorIndexToDumpIndex = (int *) realloc (colorIndexToDumpIndex,
         (maxColors+2)*sizeof(int));
   if (maxColors > 20)
   {
      charsPerPixel = 2;
      colorChar = (char *) realloc (colorChar, (maxColors*2+4)*sizeof(char));
   }
   else
   {
      charsPerPixel = 1;
      colorChar = (char *) realloc (colorChar, (maxColors+2)*sizeof(char));
   }
   colorStr = (char * *) realloc (colorStr, (maxColors+2)*sizeof(char *));
   colorStr[maxColors-1] = NULL;
   colorIndexToDumpIndex[maxColors-1] = INVALID;
}

void InitXBm ()
{
   XGCValues	values;
   char		* c_ptr;

   dummyBitmap = XCreatePixmap (mainDisplay, mainWindow, 1, 1, 1);

   values.foreground = 0;
   values.background = 0;
   values.function = GXcopy;
   values.fill_style = FillSolid;
   xbmGC = XCreateGC (mainDisplay, dummyBitmap,
         GCForeground | GCBackground | GCFunction | GCFillStyle, &values);

   importXBmRV = FALSE;
   if ((c_ptr = XGetDefault(mainDisplay,TOOL_NAME,"XBmReverseVideo")) != NULL)
      if (strcmp ("True", c_ptr) == 0 || strcmp ("true", c_ptr) == 0)
         importXBmRV = TRUE;

   askForXBmSpec = FALSE;
   if ((c_ptr = XGetDefault(mainDisplay,TOOL_NAME,"AskForXBmSpec")) != NULL)
      if (strcmp ("True", c_ptr) == 0 || strcmp ("true", c_ptr) == 0)
         askForXBmSpec = TRUE;

   stripEPSComments = TRUE;
   if ((c_ptr = XGetDefault(mainDisplay,TOOL_NAME,"StripEPSComments")) != NULL)
      if (strcmp ("False", c_ptr) == 0 || strcmp ("false", c_ptr) == 0)
         stripEPSComments = FALSE;

   xpmOutputVersion = 1;
   if ((c_ptr = XGetDefault(mainDisplay,TOOL_NAME,"XPmOutputVersion")) != NULL)
   {
      xpmOutputVersion = atoi (c_ptr);
      if (xpmOutputVersion != 1 && xpmOutputVersion != 3)
      {
         fprintf (stderr, "Invalid XPmOutputVersion (%s) in X Defaults.  %s\n",
               c_ptr, "1 is used.");
         xpmOutputVersion = 1;
      }
   }

   xpmInXGrabSCFormat = FALSE;
   if ((c_ptr = XGetDefault(mainDisplay,TOOL_NAME,"XPmInXGrabSCFormat"))!=NULL)
      if (strcmp ("True", c_ptr) == 0 || strcmp ("true", c_ptr) == 0)
         xpmInXGrabSCFormat = TRUE;

   halfToneBitmap = FALSE;
   if ((c_ptr = XGetDefault(mainDisplay,TOOL_NAME,"HalfToneBitmap"))!=NULL)
      if (strcmp ("True", c_ptr) == 0 || strcmp ("true", c_ptr) == 0)
         halfToneBitmap = TRUE;

   thresholdBitmap = FALSE;
   if ((c_ptr = XGetDefault(mainDisplay,TOOL_NAME,"ThresholdBitmap"))!=NULL)
      if (!halfToneBitmap && (strcmp ("True", c_ptr) == 0 ||
            strcmp ("true", c_ptr) == 0))
         thresholdBitmap = TRUE;

   bitmapThreshold = (halfToneBitmap) ? 0.5 : 1.0;
   strcpy (bitmapThresholdStr, ((halfToneBitmap) ? "0.5" : "1.0"));
   if ((c_ptr = XGetDefault(mainDisplay,TOOL_NAME,"BitmapThreshold"))!=NULL)
   {
      bitmapThreshold = atof (c_ptr);
      strcpy (bitmapThresholdStr, c_ptr);
      if (bitmapThreshold < 0 || bitmapThreshold > 1)
      {
         fprintf (stderr, "%s (%s) in X Defaults.  %s is used.\n",
               "Invalid BitmapThreshold", c_ptr,
               ((halfToneBitmap) ? "0.5" : "1.0"));
         bitmapThreshold = (halfToneBitmap) ? 0.5 : 1.0;
         strcpy (bitmapThresholdStr, ((halfToneBitmap) ? "0.5" : "1.0"));
      }
   }

   InitEPS ();
}

void CleanUpXBm ()
{
   register int	i;

   if (colorChar != NULL)
   {
      for (i = 0; i < numColorsToDump+2; i++)
         if (colorStr[i] != NULL)
            free (colorStr[i]);
      free (colorStr);
      free (colorChar);
      free (pixelValue);
      free (colorIndexToDumpIndex);
   }
   CleanUpEPS ();

   if (xbmGC != NULL) XFreeGC (mainDisplay, xbmGC);
   XFreePixmap (mainDisplay, dummyBitmap);

   importXBmRV = FALSE;
   askForXBmSpec = FALSE;
}

static
void MultiplyTwoByTwo (m1, m2)
   float	m1[2][2], m2[2][2];
{
   float	m3[2][2];

   m3[0][0] = m1[0][0]*m2[0][0] + m1[0][1]*m2[1][0];
   m3[0][1] = m1[0][0]*m2[0][1] + m1[0][1]*m2[1][1];
   m3[1][0] = m1[1][0]*m2[0][0] + m1[1][1]*m2[1][0];
   m3[1][1] = m1[1][0]*m2[0][1] + m1[1][1]*m2[1][1];

   m1[0][0]=m3[0][0]; m1[0][1]=m3[0][1]; m1[1][0]=m3[1][0]; m1[1][1]=m3[1][1];
}

void CalcTransform (mtrx)
   register struct MtrxRec	* mtrx;
{
   float	m1[2][2], m2[2][2], tmp_val;

   switch (mtrx->rotate)
   {
      case ROTATE0:
      case ROTATE180:
         mtrx->h_scale = mtrx->w / mtrx->image_w;
         mtrx->v_scale = mtrx->h / mtrx->image_h;
         break;
      case ROTATE90:
      case ROTATE270:
         mtrx->h_scale = mtrx->h / mtrx->image_w;
         mtrx->v_scale = mtrx->w / mtrx->image_h;
         break;
   }

   m1[0][0]=1.0; m1[0][1]=0.0; m1[1][0]=0.0; m1[1][1]=1.0;
   if (mtrx->flip & HORI_EVEN)
   {
      m2[0][0]=(-1.0); m2[0][1]=0.0; m2[1][0]=0.0; m2[1][1]=1.0;
      MultiplyTwoByTwo (m1, m2);
   }
   if (mtrx->flip & VERT_EVEN)
   {
      m2[0][0]=1.0; m2[0][1]=0.0; m2[1][0]=0.0; m2[1][1]=(-1.0);
      MultiplyTwoByTwo (m1, m2);
   }
   switch (mtrx->rotate)
   {
      case ROTATE0:
         if (mtrx->flip & (HORI_ODD | VERT_ODD))
         {
            m2[0][0]=rotatedCosine[ROTATE90]; m2[0][1]=rotatedSine[ROTATE90];
            m2[1][0]=(-rotatedSine[ROTATE90]); m2[1][1]=rotatedCosine[ROTATE90];
            MultiplyTwoByTwo (m1, m2);
            if (mtrx->flip & HORI_ODD)
            {
               m2[0][0]=(-1.0); m2[0][1]=0.0; m2[1][0]=0.0; m2[1][1]=1.0;
               MultiplyTwoByTwo (m1, m2);
            }
            if (mtrx->flip & VERT_ODD)
            {
               m2[0][0]=1.0; m2[0][1]=0.0; m2[1][0]=0.0; m2[1][1]=(-1.0);
               MultiplyTwoByTwo (m1, m2);
            }
            m2[0][0]=rotatedCosine[ROTATE270];
            m2[0][1]=rotatedSine[ROTATE270];
            m2[1][0]=(-rotatedSine[ROTATE270]);
            m2[1][1]=rotatedCosine[ROTATE270];
            MultiplyTwoByTwo (m1, m2);
         }
         break;
      case ROTATE90:
         m2[0][0]=rotatedCosine[ROTATE90]; m2[0][1]=rotatedSine[ROTATE90];
         m2[1][0]=(-rotatedSine[ROTATE90]); m2[1][1]=rotatedCosine[ROTATE90];
         MultiplyTwoByTwo (m1, m2);
         if (mtrx->flip & HORI_ODD)
         {
            m2[0][0]=(-1.0); m2[0][1]=0.0; m2[1][0]=0.0; m2[1][1]=1.0;
            MultiplyTwoByTwo (m1, m2);
         }
         if (mtrx->flip & VERT_ODD)
         {
            m2[0][0]=1.0; m2[0][1]=0.0; m2[1][0]=0.0; m2[1][1]=(-1.0);
            MultiplyTwoByTwo (m1, m2);
         }
         break;
      case ROTATE180:
         m2[0][0]=rotatedCosine[ROTATE90]; m2[0][1]=rotatedSine[ROTATE90];
         m2[1][0]=(-rotatedSine[ROTATE90]); m2[1][1]=rotatedCosine[ROTATE90];
         MultiplyTwoByTwo (m1, m2);
         if (mtrx->flip & HORI_ODD)
         {
            m2[0][0]=(-1.0); m2[0][1]=0.0; m2[1][0]=0.0; m2[1][1]=1.0;
            MultiplyTwoByTwo (m1, m2);
         }
         if (mtrx->flip & VERT_ODD)
         {
            m2[0][0]=1.0; m2[0][1]=0.0; m2[1][0]=0.0; m2[1][1]=(-1.0);
            MultiplyTwoByTwo (m1, m2);
         }
         m2[0][0]=rotatedCosine[ROTATE90]; m2[0][1]=rotatedSine[ROTATE90];
         m2[1][0]=(-rotatedSine[ROTATE90]); m2[1][1]=rotatedCosine[ROTATE90];
         MultiplyTwoByTwo (m1, m2);
         break;
      case ROTATE270:
         m2[0][0]=rotatedCosine[ROTATE270]; m2[0][1]=rotatedSine[ROTATE270];
         m2[1][0]=(-rotatedSine[ROTATE270]); m2[1][1]=rotatedCosine[ROTATE270];
         MultiplyTwoByTwo (m1, m2);
         if (mtrx->flip & HORI_ODD)
         {
            m2[0][0]=(-1.0); m2[0][1]=0.0; m2[1][0]=0.0; m2[1][1]=1.0;
            MultiplyTwoByTwo (m1, m2);
         }
         if (mtrx->flip & VERT_ODD)
         {
            m2[0][0]=1.0; m2[0][1]=0.0; m2[1][0]=0.0; m2[1][1]=(-1.0);
            MultiplyTwoByTwo (m1, m2);
         }
         break;
   }
   m2[0][0]=mtrx->h_scale; m2[0][1]=0.0; m2[1][0]=0.0; m2[1][1]=mtrx->v_scale;
   MultiplyTwoByTwo (m2, m1);

   mtrx->m[0][0]=m2[0][0]; mtrx->m[0][1]=m2[0][1];
   mtrx->m[1][0]=m2[1][0]; mtrx->m[1][1]=m2[1][1];
   tmp_val = m2[0][0]*m2[1][1] - m2[0][1]*m2[1][0];
   if (fabs((double) tmp_val) < 1.0e-6)
      tmp_val = ((float)1000000);
   else
      tmp_val = ((float)1.0)/tmp_val;
   mtrx->rev_m[0][0] = tmp_val*m2[1][1];
   mtrx->rev_m[0][1] = (-tmp_val)*m2[0][1];
   mtrx->rev_m[1][0] = (-tmp_val)*m2[1][0];
   mtrx->rev_m[1][1] = tmp_val*m2[0][0];

   mtrx->transformed_w = mtrx->image_w*m2[0][0] + mtrx->image_h*m2[1][0];
   mtrx->transformed_h = mtrx->image_w*m2[0][1] + mtrx->image_h*m2[1][1];

   switch (mtrx->rotate)
   {
      case ROTATE0:
         mtrx->dump_h_scale=mtrx->m[0][0]; mtrx->dump_v_scale=mtrx->m[1][1];
         mtrx->degree=0;
         break;
      case ROTATE90:
         mtrx->dump_h_scale=mtrx->m[1][0]; mtrx->dump_v_scale=(-mtrx->m[0][1]);
         mtrx->degree=(-90);
         break;
      case ROTATE180:
         mtrx->dump_h_scale=(-mtrx->m[0][0]);
         mtrx->dump_v_scale=(-mtrx->m[1][1]);
         mtrx->degree=(-180);
         break;
      case ROTATE270:
         mtrx->dump_h_scale=(-mtrx->m[1][0]); mtrx->dump_v_scale=mtrx->m[0][1];
         mtrx->degree=(-270);
         break;
   }
}

void MakeCachedBitmap (ObjPtr)
   struct ObjRec	* ObjPtr;
{
   register int		r, c;
   int			w, h, rotate, flip;
   int			num_cols, num_rows, image_w, image_h, watch_cursor;
   int			start_col, start_row;
   struct XBmRec	* xbm_ptr = ObjPtr->detail.xbm;
   struct MtrxRec	mtrx;
   Pixmap		dest_bitmap;
   XImage		* image, * dest_image;

   if (xbm_ptr->real_type==XBM_EPS && xbm_ptr->bitmap==None) return;

   w = ObjPtr->obbox.rbx - ObjPtr->obbox.ltx;
   h = ObjPtr->obbox.rby - ObjPtr->obbox.lty;
   num_cols = (zoomedIn) ? (w<<zoomScale) : (w>>zoomScale);
   num_rows = (zoomedIn) ? (h<<zoomScale) : (h>>zoomScale);

   if (xbm_ptr->cached_bitmap!=None &&
         xbm_ptr->cached_zoomed==zoomedIn && xbm_ptr->cached_zoom==zoomScale &&
         xbm_ptr->cached_w==num_cols && xbm_ptr->cached_h==num_rows &&
         xbm_ptr->cached_rotate==xbm_ptr->rotate &&
         xbm_ptr->cached_flip==xbm_ptr->flip)
      return;

   if ((w>>zoomScale)==0 || (h>>zoomScale)==0)
   {
      if (xbm_ptr->cached_bitmap != None)
         XFreePixmap (mainDisplay, xbm_ptr->cached_bitmap);
      xbm_ptr->cached_bitmap = None;
      return;
   }

   watch_cursor = watchCursorOnMainWindow;
   if (!watch_cursor)
   {
      SetWatchCursor (drawWindow);
      SetWatchCursor (mainWindow);
   }

   image = xbm_ptr->image;
   flip = xbm_ptr->flip;
   rotate = xbm_ptr->rotate;
   image_w = xbm_ptr->image_w;
   image_h = xbm_ptr->image_h;
   if (xbm_ptr->cached_bitmap != None)
      XFreePixmap (mainDisplay, xbm_ptr->cached_bitmap);
   xbm_ptr->cached_bitmap = None;

   if (image == NULL)
      image = xbm_ptr->image = XGetImage (mainDisplay, xbm_ptr->bitmap, 0, 0,
            image_w, image_h, 1, ZPixmap);

   mtrx.image_w = (float)image_w; mtrx.image_h = (float)image_h;
   mtrx.w = (float)num_cols; mtrx.h = (float)num_rows;
   mtrx.rotate = rotate; mtrx.flip = flip;

   CalcTransform (&mtrx);

   start_col = (mtrx.transformed_w >= 0.0) ? 0 : (-num_cols)+1;
   start_row = (mtrx.transformed_h >= 0.0) ? 0 : (-num_rows)+1;

   dest_bitmap = XCreatePixmap (mainDisplay,dummyBitmap,num_cols,num_rows,1);
   XFillRectangle (mainDisplay,dest_bitmap,xbmGC,0,0,num_cols,num_rows);
   dest_image = XGetImage (mainDisplay, dest_bitmap, 0, 0, num_cols,
         num_rows, 1, ZPixmap);

   for (r = 0; r < num_rows; r++)
   {
      float	part_x, part_y;

      part_x = (r+start_row)*mtrx.rev_m[1][0];
      part_y = (r+start_row)*mtrx.rev_m[1][1];
      for (c = 0; c < num_cols; c++)
      {
         int	x, y;

         x = (int)((c+start_col)*mtrx.rev_m[0][0]+part_x);
         y = (int)((c+start_col)*mtrx.rev_m[0][1]+part_y);
         if (x>=0 && x<image_w && y>=0 && y<image_h && XGetPixel(image,x,y)==1)
            XPutPixel (dest_image, c, r, 1);
      }
   }

   XPutImage (mainDisplay, dest_bitmap, xbmGC, dest_image, 0, 0, 0, 0,
         num_cols, num_rows);
   xbm_ptr->cached_bitmap = dest_bitmap;
   xbm_ptr->cached_zoomed = zoomedIn;
   xbm_ptr->cached_zoom = zoomScale;
   xbm_ptr->cached_rotate = xbm_ptr->rotate;
   xbm_ptr->cached_flip = xbm_ptr->flip;
   xbm_ptr->cached_w = num_cols;
   xbm_ptr->cached_h = num_rows;

   XDestroyImage (dest_image);

   if (!watch_cursor)
   {
      SetDefaultCursor (mainWindow);
      ShowCursor ();
   }
}

int ExtractBitmap (OrigBitmap, OrigImage, X, Y, W, H, Bitmap, Image)
   Pixmap	OrigBitmap, * Bitmap;
   XImage	* OrigImage, * * Image;
   int		X, Y, W, H;
{
   register int	j, i;
   char		msg[MAXSTRING];
   XImage	* src_image;

   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);

   if ((*Bitmap=XCreatePixmap (mainDisplay,mainWindow,W,H,1)) == None)
   {
      sprintf (msg, "Can not allocate bitmap of size %1dx%1d.", W, H);
      Msg (msg);
      SetDefaultCursor (mainWindow);
      SetDefaultCursor (drawWindow);
      return (FALSE);
   }

   if ((*Image = XGetImage (mainDisplay, *Bitmap, 0, 0, W, H, 1,
         ZPixmap)) == NULL)
   {
      Msg ("XGetImage() failed!  May have run out of memory!");
      XFreePixmap (mainDisplay, *Bitmap); *Bitmap = None;
      SetDefaultCursor (mainWindow);
      SetDefaultCursor (drawWindow);
      return (FALSE);
   }

   if (OrigImage!=NULL && X==0 && Y==0)
      src_image = OrigImage;
   else
   {
      if ((src_image = XGetImage (mainDisplay, OrigBitmap, X, Y, W, H, 1,
            ZPixmap)) == NULL)
      {
         Msg ("XGetImage() failed!  May have run out of memory!");
         XFreePixmap (mainDisplay, *Bitmap); *Bitmap = None;
         XDestroyImage (*Image); *Image = NULL;
         SetDefaultCursor (mainWindow);
         SetDefaultCursor (drawWindow);
         return (FALSE);
      }
   }

   for (i = 0; i < H; i++)
      for (j = 0; j < W; j++)
         XPutPixel (*Image, j, i, XGetPixel (src_image, j, i));

   XPutImage (mainDisplay, *Bitmap, xbmGC, *Image, 0, 0, 0, 0, W, H);
   SetDefaultCursor (mainWindow);
   SetDefaultCursor (drawWindow);

   return (TRUE);
}

static
void InvertXBmObject (ObjPtr)
   struct ObjRec	* ObjPtr;
{
   register int		i, j;
   int			image_w, image_h, pixel;
   Pixmap		bitmap;
   XImage		* image;

   SetWatchCursor (drawWindow);
   SetWatchCursor (mainWindow);

   bitmap = ObjPtr->detail.xbm->bitmap;

   image_w = ObjPtr->detail.xbm->image_w;
   image_h = ObjPtr->detail.xbm->image_h;
   if (ObjPtr->detail.xbm->image == NULL)
   {
      if ((image = ObjPtr->detail.xbm->image = XGetImage (mainDisplay, bitmap,
            0, 0, image_w, image_h, 1, ZPixmap)) == NULL)
      {
         Msg ("XGetImage() failed!  May have run out of memory!");
         SetDefaultCursor (mainWindow);
         SetDefaultCursor (drawWindow);
         return;
      }
   }
   else
      image = ObjPtr->detail.xbm->image;

   for (i = 0; i < image_h; i++)
      for (j = 0; j < image_w; j++)
      {
         pixel = XGetPixel (image, j, i);
         XPutPixel (image, j, i, ((pixel==1) ? 0 : 1));
      }
   XPutImage (mainDisplay, bitmap, xbmGC, image, 0, 0, 0, 0, image_w, image_h);

   if (ObjPtr->detail.xbm->cached_bitmap != None)
      XFreePixmap (mainDisplay, ObjPtr->detail.xbm->cached_bitmap);
   ObjPtr->detail.xbm->cached_bitmap = None;
   if (zoomScale != 0)
      ObjPtr->detail.xbm->cached_zoom = 0;
   else
      ObjPtr->detail.xbm->cached_rotate = INVALID;

   SetDefaultCursor (mainWindow);
   SetDefaultCursor (drawWindow);
}

static
int ObjListInvertable (LastObjPtr)
   struct ObjRec	* LastObjPtr;
{
   register struct ObjRec	* obj_ptr;

   for (obj_ptr = LastObjPtr; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
      switch (obj_ptr->type)
      {
         case OBJ_XBM: return (TRUE);

         case OBJ_SYM:
         case OBJ_GROUP:
            if (ObjListInvertable (obj_ptr->detail.r->last)) return (TRUE);
            break;
      }
   return (FALSE);
}

static
void InvertObjListXBitmap (LastObjPtr)
   struct ObjRec	* LastObjPtr;
{
   register struct ObjRec	* obj_ptr;

   for (obj_ptr = LastObjPtr; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
      switch (obj_ptr->type)
      {
         case OBJ_XBM: InvertXBmObject (obj_ptr); break;

         case OBJ_SYM:
         case OBJ_GROUP: InvertObjListXBitmap (obj_ptr->detail.r->last); break;
      }
}

void InvertXBitmaps ()
{
   register struct SelRec	* sel_ptr;
   int				changed = FALSE;

   if (topSel == NULL) return;

   StartCompositeCmd ();
   for (sel_ptr = botSel; sel_ptr != NULL; sel_ptr = sel_ptr->prev)
      switch (sel_ptr->obj->type)
      {
         case OBJ_XBM:
            changed = TRUE;
            PrepareToReplaceAnObj (sel_ptr->obj);
            InvertXBmObject (sel_ptr->obj);
            RecordReplaceAnObj (sel_ptr->obj);
            break;

         case OBJ_SYM:
         case OBJ_GROUP:
            if (ObjListInvertable (sel_ptr->obj->detail.r->last))
            {
               changed = TRUE;
               PrepareToReplaceAnObj (sel_ptr->obj);
               InvertObjListXBitmap (sel_ptr->obj->detail.r->last);
               RecordReplaceAnObj (sel_ptr->obj);
            }
            break;
      }
   EndCompositeCmd ();

   if (changed)
   {
      SetFileModified (TRUE);
      HighLightReverse ();
      RedrawAnArea (botObj, selLtX-GRID_ABS_SIZE(1), selLtY-GRID_ABS_SIZE(1),
            selRbX+GRID_ABS_SIZE(1), selRbY+GRID_ABS_SIZE(1));
      HighLightForward ();
   }
}

static
void ParseGeomSpec (geom_spec, image_w, image_h, src_x, src_y, src_w, src_h)
   char		* geom_spec;
   int		image_w, image_h, * src_x, * src_y, * src_w, * src_h;
{
   int		bitmask;
   XSizeHints	sizehints;

   *src_x = *src_y = 0;
   *src_w = image_w;
   *src_h = image_h;

   bitmask = XParseGeometry (geom_spec, &(sizehints.x), &(sizehints.y),
         (unsigned int *)&(sizehints.width),
         (unsigned int *)&(sizehints.height));
   if (bitmask & WidthValue) *src_w = sizehints.width;
   if (bitmask & HeightValue) *src_h = sizehints.height;
   if (bitmask & XValue)
   {
      *src_x = (bitmask & XNegative) ? image_w+sizehints.x : sizehints.x;
      if (bitmask & XNegative)
      {
         if (*src_w > image_w) *src_w = image_w;
         if (*src_x >= image_w || *src_x < 0) *src_x = image_w;
         if (*src_x-*src_w < 0)
         {
            *src_w = *src_x;
            *src_x = 0;
         }
         else
            *src_x -= *src_w;
      }
      else
      {
         if (*src_w > image_w) *src_w = image_w;
         if (*src_x >= image_w || *src_x < 0) *src_x = 0;
         if (*src_x+*src_w > image_w) *src_w = image_w-*src_x;
      }
   }
   if (bitmask & YValue)
   {
      *src_y = (bitmask & YNegative) ? image_h+sizehints.y : sizehints.y;
      if (bitmask & YNegative)
      {
         if (*src_h > image_h) *src_h = image_h;
         if (*src_y >= image_h || *src_y < 0) *src_y = image_h;
         if (*src_y-*src_h < 0)
         {
            *src_h = *src_y;
            *src_y = 0;
         }
         else
            *src_y -= *src_h;
      }
      else
      {
         if (*src_h > image_h) *src_h = image_h;
         if (*src_y >= image_h || *src_y < 0) *src_y = 0;
         if (*src_y+*src_h > image_h) *src_h = image_h-*src_y;
      }
   }
   if ((bitmask&WidthValue) && *src_x+*src_w>image_w) *src_w = image_w-*src_x;
   if ((bitmask&HeightValue) && *src_y+*src_h>image_h) *src_h = image_h-*src_y;
}

static
char * FindEqual(s)
   register char	* s;
{
   while (*s != '=' && *s != '\0') s++;
   return ((*s == '=') ? (s) : (char *)NULL);
}

void ParseCutSpec (spec, image_w, image_h, mag, src_x, src_y, src_w, src_h)
   char		* spec;
   int		image_w, image_h, * src_x, * src_y, * src_w, * src_h;
   float	* mag;
{
   char	* geom_spec;

   *mag = 1.0;

   if ((geom_spec = FindEqual (spec)) == NULL)
      ParseGeomSpec (spec,image_w,image_h,src_x,src_y,src_w,src_h);
   else
   {
      *geom_spec = '\0';
      geom_spec++;
      ParseGeomSpec (geom_spec,image_w,image_h,src_x,src_y,src_w,src_h);

      sscanf (spec, "%f", mag);
      if (*mag <= 0.0) *mag = 1.0;
   }
}

void CutXBitmap ()
{
   int			w, h;
   int			ltx, lty, rbx, rby, new_w, new_h;
   int			src_x, src_y, src_w, src_h, image_w, image_h;
   char			mag_spec[MAXSTRING], msg[MAXSTRING];
   float		h_scale=1.0, v_scale=1.0, mag;
   Pixmap		dest_bitmap=None;
   XImage		* dest_image=NULL;
   struct ObjRec	* obj_ptr = topSel->obj, * new_obj_ptr;
   struct XBmRec	* new_xbm_ptr;

   if (obj_ptr->detail.xbm->real_type==XBM_EPS)
   {
      Msg ("Can not cut an EPS object.");
      return;
   }

   src_x = 0;
   src_y = 0;
   src_w = image_w = obj_ptr->detail.xbm->image_w;
   src_h = image_h = obj_ptr->detail.xbm->image_h;
   mag = 1.0;

   switch (obj_ptr->detail.xbm->rotate)
   {
      case ROTATE0:
      case ROTATE180:
         h_scale = ((float)((float)(obj_ptr->obbox.rbx-obj_ptr->obbox.ltx)) /
               ((float)image_w));
         v_scale = ((float)((float)(obj_ptr->obbox.rby-obj_ptr->obbox.lty)) /
               ((float)image_h));
         break;
      case ROTATE90:
      case ROTATE270:
         h_scale = ((float)((float)(obj_ptr->obbox.rby-obj_ptr->obbox.lty)) /
               ((float)image_w));
         v_scale = ((float)((float)(obj_ptr->obbox.rbx-obj_ptr->obbox.ltx)) /
               ((float)image_h));
         break;
   }

   sprintf (msg, "%s: [[MAG=]WxH+X+Y] (original size is %1dx%1d)",
         "Please enter geometry spec", image_w, image_h);
   Dialog (msg, "( <CR>: accept, <ESC>: cancel )", mag_spec);
   if (*mag_spec == '\0') return;

   ParseCutSpec (mag_spec,image_w,image_h,&mag,&src_x,&src_y,&src_w,&src_h);
   if (src_x==0 && src_y==0 && src_w==image_w && src_h==image_h && mag==1.0)
      return;

   if (src_w==0 || src_h==0)
   {
      Msg ("Bitmap can not have 0 width or height.");
      return;
   }

   PrepareToRecord (CMD_REPLACE, topSel, botSel, numObjSelected);
   if (!ExtractBitmap (obj_ptr->detail.xbm->bitmap,
         obj_ptr->detail.xbm->image, src_x, src_y, src_w, src_h,
         &dest_bitmap, &dest_image))
   {
      AbortPrepareCmd (CMD_REPLACE);
      return;
   }

   sprintf (msg, "New bitmap size is %1dx%1d.", src_w, src_h);
   Msg (msg);

   UnlinkObj (obj_ptr);

   ltx = selLtX; lty = selLtY; rbx = selRbX; rby = selRbY;
   HighLightReverse ();

   w = new_w = (int)(((float)src_w) * mag);
   h = new_h = (int)(((float)src_h) * mag);

   new_obj_ptr = (struct ObjRec *) calloc (1, sizeof(struct ObjRec));
   DupObjBasics (obj_ptr, new_obj_ptr);

   new_xbm_ptr = (struct XBmRec *) calloc (1, sizeof(struct XBmRec));
   new_obj_ptr->detail.xbm = new_xbm_ptr;

   new_xbm_ptr->image = dest_image;
   new_xbm_ptr->image_w = src_w;
   new_xbm_ptr->image_h = src_h;
   new_xbm_ptr->bitmap = dest_bitmap;
   new_xbm_ptr->data = NULL;
   new_xbm_ptr->fill = obj_ptr->detail.xbm->fill;
   new_xbm_ptr->rotate = obj_ptr->detail.xbm->rotate;
   new_xbm_ptr->flip = obj_ptr->detail.xbm->flip;
   new_xbm_ptr->cached_zoom = 0;
   new_xbm_ptr->cached_bitmap = None;
   new_xbm_ptr->cached_rotate = INVALID;
   new_xbm_ptr->cached_flip = 0;
   new_xbm_ptr->cached_w = 0;
   new_xbm_ptr->cached_h = 0;

   switch (obj_ptr->detail.xbm->rotate)
   {
      case ROTATE0:
      case ROTATE180:
         new_w = round(h_scale * ((float)w));
         new_h = round(v_scale * ((float)h));
         break;
      case ROTATE90:
      case ROTATE270:
         new_h = round(h_scale * ((float)w));
         new_w = round(v_scale * ((float)h));
         break;
   }

   new_obj_ptr->obbox.ltx = obj_ptr->obbox.ltx;
   new_obj_ptr->obbox.lty = obj_ptr->obbox.lty;
   new_obj_ptr->obbox.rbx = new_obj_ptr->bbox.rbx = obj_ptr->obbox.ltx+new_w;
   new_obj_ptr->obbox.rby = new_obj_ptr->bbox.rby = obj_ptr->obbox.lty+new_h;

   AdjObjBBox (new_obj_ptr);

   topSel->obj = botSel->obj = new_obj_ptr;
   AddObj (NULL, topObj, new_obj_ptr);
   RecordCmd (CMD_REPLACE, NULL, topSel, botSel, numObjSelected);
   FreeObj (obj_ptr);

   UpdSelBBox ();
   RedrawAreas (botObj, ltx-GRID_ABS_SIZE(1), lty-GRID_ABS_SIZE(1),
         rbx+GRID_ABS_SIZE(1), rby+GRID_ABS_SIZE(1),
         selLtX-GRID_ABS_SIZE(1), selLtY-GRID_ABS_SIZE(1),
         selRbX+GRID_ABS_SIZE(1), selRbY+GRID_ABS_SIZE(1));
   HighLightForward ();
   SetFileModified (TRUE);
   justDupped = FALSE;
}

static
void BuildObjXPmColors (ObjPtr)
   struct ObjRec	* ObjPtr;
{
   register int			i, color_index;
   register struct ObjRec	* obj_ptr;
   register struct AttrRec	* attr_ptr;
   struct XPmRec		* xpm_ptr;
   int				start_index, new_alloc;

   switch (ObjPtr->type)
   {
      case OBJ_POLY:
      case OBJ_BOX:
      case OBJ_OVAL:
      case OBJ_TEXT:
      case OBJ_POLYGON:
      case OBJ_ARC:
      case OBJ_RCBOX:
      case OBJ_XBM:
         color_index = ObjPtr->color;
         if (colorIndexToDumpIndex[color_index] == INVALID)
         {
            if (colorStr[numColorsToDump] != NULL)
               free (colorStr[numColorsToDump]);

            pixelValue[numColorsToDump] = colorPixels[color_index];
            colorIndexToDumpIndex[color_index] = numColorsToDump;
            colorStr[numColorsToDump] = (char *) calloc
                  (strlen(colorMenuItems[color_index])+1, sizeof(char));
            strcpy (colorStr[numColorsToDump], colorMenuItems[color_index]);
            numColorsToDump++;
         }
         break;

      case OBJ_XPM:
         xpm_ptr = ObjPtr->detail.xpm;
         start_index = (xpm_ptr->first_pixel_is_bg) ? 1 : 0;
         for (i = start_index; i < xpm_ptr->ncolors; i++)
         {
            if ((color_index =
                  QuickFindColorIndex (xpm_ptr->color_str[i], &new_alloc)) !=
                  INVALID)
            {
               if (new_alloc) ExpandTmpStorage ();
               if (colorIndexToDumpIndex[color_index] == INVALID)
               {
                  if (colorStr[numColorsToDump] != NULL)
                     free (colorStr[numColorsToDump]);

                  pixelValue[numColorsToDump] = colorPixels[color_index];
                  colorIndexToDumpIndex[color_index] = numColorsToDump;
                  colorStr[numColorsToDump] = (char *) calloc
                        (strlen(colorMenuItems[color_index])+1, sizeof(char));
                  strcpy (colorStr[numColorsToDump],
                        colorMenuItems[color_index]);
                  numColorsToDump++;
               }
            }
         }
         break;

      case OBJ_GROUP:
      case OBJ_SYM:
      case OBJ_ICON:
         for (obj_ptr = ObjPtr->detail.r->first; obj_ptr != NULL;
               obj_ptr = obj_ptr->next)
            BuildObjXPmColors (obj_ptr);
         break;
   }
   for (attr_ptr=ObjPtr->fattr; attr_ptr!=NULL; attr_ptr=attr_ptr->next)
   {
      color_index = attr_ptr->obj->color;
      if (colorIndexToDumpIndex[color_index] == INVALID)
      {
         if (colorStr[numColorsToDump] != NULL)
            free (colorStr[numColorsToDump]);

         pixelValue[numColorsToDump] = colorPixels[color_index];
         colorIndexToDumpIndex[color_index] = numColorsToDump;
         colorStr[numColorsToDump] = (char *) calloc
               (strlen(colorMenuItems[color_index])+1, sizeof(char));
         strcpy (colorStr[numColorsToDump], colorMenuItems[color_index]);
         numColorsToDump++;
      }
   }
}

static
void BuildXPmColors ()
{
   register int			i;
   register struct ObjRec	* obj_ptr;

   if (colorChar == NULL)
   {
      pixelValue = (int *) calloc (maxColors+2, sizeof(int));
      colorIndexToDumpIndex = (int *) calloc (maxColors+2, sizeof(int));
      if (maxColors > 20)
      {
         charsPerPixel = 2;
         colorChar = (char *) calloc (maxColors*2+4, sizeof(char));
      }
      else
      {
         charsPerPixel = 1;
         colorChar = (char *) calloc (maxColors+2, sizeof(char));
      }
      colorStr = (char * *) calloc (maxColors+2, sizeof(char *));

      for (i = 0; i < maxColors+2; i++) colorStr[i] = NULL;
   }

   for (i = 0; i < maxColors; i++) colorIndexToDumpIndex[i] = INVALID;

   pixelValue[0] = myBgPixel;
   colorStr[0] = (char *) calloc (strlen(myBgColorStr)+1, sizeof(char));
   strcpy (colorStr[0], myBgColorStr);
   numColorsToDump = 1;

   for (obj_ptr = botObj; obj_ptr != NULL; obj_ptr = obj_ptr->prev)
      BuildObjXPmColors (obj_ptr);

   colorChar[0] = '`';
   if (charsPerPixel > 1) colorChar[1] = '`';
   for (i = 1; i < numColorsToDump; i++)
   {
      if (charsPerPixel == 1)
         colorChar[i] = (char)(((int)('a'))+i-1);
      else
      {
         colorChar[i*2] = (char)(((int)('a'))+(int)(i/10));
         colorChar[i*2+1] = (char)(((int)('0'))+(i%10));
      }
   }
}

static
void ColorStrToXPmStr (index, color_str)
   int	index;
   char	* color_str;
{
   char	s[3];
   int	i, value;

   if (*(colorStr[index]) == '#')
      strcpy (color_str, colorStr[index]);
   else
   {
      for (i = 0; i < maxColors; i++)
         if (colorPixels[i] ==  pixelValue[index])
            break;

      if (i == maxColors)
         strcpy (color_str, colorStr[index]);
      else
      {
         strcpy (color_str, "#");
         value = (int) (((float)tgifColors[i].red/maxRGB) * 0x100);
         if (value > 255)
            value = 255;
         else if (value < 255)
            value = 0;
         sprintf (s, "%c%c", hexValue[(value>>4)&0x0f], hexValue[value&0x0f]);
         strcat (color_str, s);
         strcat (color_str, s);

         value = (int) (((float)tgifColors[i].green/maxRGB) * 0x100);
         if (value > 255)
            value = 255;
         else if (value < 255)
            value = 0;
         sprintf (s, "%c%c", hexValue[(value>>4)&0x0f], hexValue[value&0x0f]);
         strcat (color_str, s);
         strcat (color_str, s);

         value = (int) (((float)tgifColors[i].blue/maxRGB) * 0x100);
         if (value > 255)
            value = 255;
         else if (value < 255)
            value = 0;
         sprintf (s, "%c%c", hexValue[(value>>4)&0x0f], hexValue[value&0x0f]);
         strcat (color_str, s);
         strcat (color_str, s);
      }
   }
}

static
void DumpXPmColors (FP)
   FILE	* FP;
{
   register int	i, j;
   char		s[MAXSTRING];

   if (xpmOutputVersion == 1)
   {
      if (xpmInXGrabSCFormat)
      {
         for (i = 0; i < numColorsToDump-1; i++)
         {
            fprintf (FP, "\"");
            for (j = 0; j < charsPerPixel; j++)
               fprintf (FP, "%c", colorChar[i*charsPerPixel+j]);
            ColorStrToXPmStr (i, s);
            fprintf (FP, "\", \"%s\"\n", s);
         }
         fprintf (FP, "\"");
         for (j = 0; j < charsPerPixel; j++)
            fprintf (FP, "%c", colorChar[i*charsPerPixel+j]);
         ColorStrToXPmStr (i, s);
         fprintf (FP, "\", \"%s\"\n} ;\n", s);
      }
      else
      {
         for (i = 0; i < numColorsToDump-1; i++)
         {
            fprintf (FP, "   \"");
            for (j = 0; j < charsPerPixel; j++)
               fprintf (FP, "%c", colorChar[i*charsPerPixel+j]);
            fprintf (FP, "\", \"%s\",\n", colorStr[i]);
         }
         fprintf (FP, "   \"");
         for (j = 0; j < charsPerPixel; j++)
            fprintf (FP, "%c", colorChar[i*charsPerPixel+j]);
         fprintf (FP, "\", \"%s\"\n};\n", colorStr[i]);
      }
   }
   else
   {  /* xpmOutputVersion is 3 */
      for (i = 0; i < numColorsToDump; i++)
      {
         fprintf (FP, "\"");
         for (j = 0; j < charsPerPixel; j++)
            fprintf (FP, "%c", colorChar[i*charsPerPixel+j]);
         fprintf (FP, " c %s\",\n", colorStr[i]);
      }
   }
}

#define FS_SCALE 0x400
#define HALF_FS_SCALE 0x200

void DumpXBitmapFile ()
{  /* called when printing in the xbm or xpm format */
   register int	j, k, bit_count, index, data;
   int		i, len, ltx, lty, w, h, byte_count, short_name, pixel;
   char		xbm_file_name[MAXPATHLENGTH], msg[MAXSTRING], * rest;
   char		name[MAXPATHLENGTH];
   FILE		* fp;
   Pixmap	pixmap;
   XImage	* image;

   strcpy (name, curFileName);
   len = strlen (name);
   for (j = len-1; name[j] != '.'; j--) ;
   name[j] = '\0';

   sprintf (xbm_file_name, "%s/%s", curDir, curFileName);
   len = strlen (xbm_file_name);
   for (j = len-1; xbm_file_name[j] != '.'; j--) ;
   if (colorDump)
      sprintf (&xbm_file_name[j], ".%s", XPM_FILE_EXT);
   else
      sprintf (&xbm_file_name[j], ".%s", XBM_FILE_EXT);

   ModifyOutputFileName (xbm_file_name);
   if (short_name = IsPrefix (bootDir, xbm_file_name, &rest)) ++rest;
   if ((fp = fopen (xbm_file_name, "w")) == NULL)
   {
      if (short_name)
         sprintf (msg, "Can not open '%s', print aborted.", rest);
      else
         sprintf (msg, "Can not open '%s', print aborted.", xbm_file_name);
      Msg (msg);
      return;
   }
   Msg ("Generating image...");

   if ((pixmap = DrawAllOnPixmap (&ltx, &lty, &w, &h)) == None) return;

   if (short_name)
      sprintf (msg, "Printing into '%s'...", rest);
   else
      sprintf (msg, "Printing into '%s'...", xbm_file_name);
   Msg (msg);

   if (colorDump)
   {
      BuildXPmColors ();
      if (xpmOutputVersion == 1)
      {
         fprintf (fp, "#define %s_format 1\n", name);
         fprintf (fp, "#define %s_width %1d\n", name, w);
         fprintf (fp, "#define %s_height %1d\n", name, h);
         fprintf (fp, "#define %s_ncolors %1d\n", name, numColorsToDump);
         fprintf (fp, "#define %s_chars_per_pixel %1d\n", name, charsPerPixel);
         if (xpmInXGrabSCFormat)
            fprintf (fp, "static char * %s_colors[] = {\n", name);
         else
            fprintf (fp, "static char *%s_colors[] = {\n", name);
         DumpXPmColors (fp);
         if (xpmInXGrabSCFormat)
            fprintf (fp, "static char * %s_pixels[] = {\n", name);
         else
            fprintf (fp, "static char *%s_pixels[] = {\n", name);
      }
      else
      {  /* xpmOutputVersion is 3 */
         fprintf (fp, "/* XPM */\n");
         fprintf (fp, "static char * %s[] = {\n", name);
         fprintf (fp, "\"%1d %1d %1d %1d\",\n", w, h, numColorsToDump,
               charsPerPixel);
         DumpXPmColors (fp);
      }
   }
   else
   {
      fprintf (fp, "#define %s_width %1d\n", name, w);
      fprintf (fp, "#define %s_height %1d\n", name, h);
      fprintf (fp, "#define %s_x_hot 0\n", name);
      fprintf (fp, "#define %s_y_hot 0\n", name);
      fprintf (fp, "static char %s_bits[] = {\n   ", name);
   }
   image = XGetImage (mainDisplay, pixmap, 0, 0, w, h, AllPlanes, ZPixmap);

   if (image == NULL)
   {
      Msg ("Can not generate image, print aborted!");
      fclose (fp);
      XFreePixmap (mainDisplay, pixmap);
      return;
   }

   if (colorDump)
   {
      BuildXPmBuckets (numColorsToDump, pixelValue, INVALID, NULL);
      for (i = 0; i < h; i++)
      {
         fprintf (fp, "\"");
         for (j = 0; j < w; j++)
         {
            data = XGetPixel (image,j,i);
            if (data == myBgPixel)
            {
               switch (charsPerPixel)
               {
                  case 1: fprintf (fp, "`"); break;
                  case 2: fprintf (fp, "``"); break;
               }
            }
            else
            {
               index = XPmLookUp (data, INVALID, NULL);
               if (index == INVALID)
               {
                  sprintf (msg, "Unrecognized pixel value %1d!  %s!",
                        data, "Print aborted");
                  Msg (msg);

                  fclose (fp);
                  XDestroyImage (image);
                  XFreePixmap (mainDisplay, pixmap);
                  return;
               }
               for (k = 0; k < charsPerPixel; k++)
                  fprintf (fp, "%c", colorChar[index*charsPerPixel+k]);
            }
         }
         if (i == h-1)
         {
            if (xpmInXGrabSCFormat)
               fprintf (fp, "\",\n} ;\n");
            else
               fprintf (fp, "\"\n};\n");
         }
         else
            fprintf (fp, "\",\n");
      }
   }
   else
   {
      if (halfToneBitmap)
      {
         /*--------------------------------------------------*/
         /* the halftoning code is adapted from 'pgmopbm.c', */
         /*       which is part of the pbmplus package.      */
         /*                                                  */
         /*       Copyright (C) 1989 by Jef Poskanzer        */
         /*--------------------------------------------------*/
         long	* thiserr, * nexterr, * tmperr, threshval, sum;
         int	fs_forward=TRUE, col, limitcol, found_index, * bits;
         float	gray;

         sprintf (msg, "    FS halftone bitmap (threshold=%s) ...",
               bitmapThresholdStr);
         Msg (msg);

         srand (0);
         thiserr = (long *) calloc (w+2, sizeof(long));
         nexterr = (long *) calloc (w+2, sizeof(long));
         bits = (int *) calloc (w+2, sizeof(long));
         for (j = 0; j < w+2; j++)
            thiserr[j] = (rand() % FS_SCALE - HALF_FS_SCALE) >> 2;
         threshval = (long)(bitmapThreshold * (float)FS_SCALE);

         byte_count = 0;
         for (i = 0; i < h; i++)
         {
            for (col = 0; col < w+2; col++) { nexterr[col] = 0; bits[col] = 0; }
            bit_count = 0;
            col = (fs_forward) ? 0 : w-1;
            limitcol = (fs_forward) ? w : -1;
            do
            {
               found_index = INVALID;
               pixel = XGetPixel (image, col, i);

               for (k = 0; k < maxColors; k++)
                  if (colorPixels[k] == pixel)
                  {
                     found_index = k;
                     break;
                  }
               if (found_index == INVALID)
               {
                  if (pixel == myBgPixel)
                     sprintf (msg, "Unrecognized pixel value %1d!  0 used.",
                           pixel);
                  else
                  {
                     sprintf (msg, "Unrecognized pixel value %1d!  1 used.",
                           pixel);
                  }
                  Msg (msg);
               }
               else
               {
                  gray = 0.299*((float)tgifColors[found_index].red/maxRGB) +
                        0.587*((float)tgifColors[found_index].green/maxRGB) +
                        0.114*((float)tgifColors[found_index].blue/maxRGB);
                  sum = ((long)(gray * FS_SCALE)) + thiserr[col+1];
                  if (sum >= threshval)
                     sum = sum-threshval-HALF_FS_SCALE;
                  else
                     bits[col] = 1; /* black bit */

                  if (fs_forward)
                  {
                     thiserr[col+2] += (sum*7)>>4;
                     nexterr[col  ] += (sum*3)>>4;
                     nexterr[col+1] += (sum*5)>>4;
                     nexterr[col+2] += (sum  )>>4;
                     col++;
                  }
                  else
                  {
                     thiserr[col  ] += (sum*7)>>4;
                     nexterr[col+2] += (sum*3)>>4;
                     nexterr[col+1] += (sum*5)>>4;
                     nexterr[col  ] += (sum  )>>4;
                     col--;
                  }
               }
            }
            while (col != limitcol);

            tmperr = thiserr;
            thiserr = nexterr;
            nexterr = tmperr;
            fs_forward = !fs_forward;

            bit_count = 0;
            data = 0;

            for (j = 0; j < w; j++)
            {
               if (bits[j]) data |= (1<<bit_count);

               if (++bit_count == 8)
               {
                  if (byte_count++ == 12)
                  {
                     byte_count = 1;
                     fprintf (fp, "\n   ");
                  }
                  fprintf (fp, "0x%c", hexValue[(data>>4) & 0xf]);

                  if (i == h-1 && j == w-1)
                     fprintf (fp, "%c};\n", hexValue[data & 0xf]);
                  else
                     fprintf (fp, "%c, ", hexValue[data & 0xf]);

                  bit_count = 0;
                  data = 0;
               }
            }
            if ((w % 8) != 0)
            {
               if (byte_count++ == 12)
               {
                  byte_count = 1;
                  fprintf (fp, "\n   ");
               }
               for (j = (w%8); j < 8; j++) data |= (1<<j);

               fprintf (fp, "0x%c", hexValue[(data>>4) & 0xf]);
               if (i == h-1)
                  fprintf (fp, "%c};\n", hexValue[data & 0xf]);
               else
                  fprintf (fp, "%c, ", hexValue[data & 0xf]);
            }
         }
      }
      else
      {
         if (thresholdBitmap)
         {
            sprintf (msg, "    Threshold bitmap (threshold=%s) ...",
                  bitmapThresholdStr);
            Msg (msg);
         }
         byte_count = 0;
         for (i = 0; i < h; i++)
         {
            bit_count = 0;
            data = 0;

            for (j = 0; j < w; j++)
            {
               if (thresholdBitmap)
               {
                  int	found_index=INVALID;
                  float	gray;

                  pixel = XGetPixel (image, j, i);
                  for (k = 0; k < maxColors; k++)
                     if (colorPixels[k] == pixel)
                     {
                        found_index = k;
                        break;
                     }
                  if (found_index == INVALID)
                  {
                     if (pixel == myBgPixel)
                        sprintf (msg, "Unrecognized pixel value %1d!  0 used.",
                              pixel);
                     else
                     {
                        sprintf (msg, "Unrecognized pixel value %1d!  1 used.",
                              pixel);
                        data |= (1<<bit_count);
                     }
                     Msg (msg);
                  }
                  else
                  {
                     gray = 0.299*((float)tgifColors[found_index].red/maxRGB) +
                           0.587*((float)tgifColors[found_index].green/maxRGB) +
                           0.114*((float)tgifColors[found_index].blue/maxRGB);
                     if (gray < bitmapThreshold)
                        data |= (1<<bit_count);
                  }
               }
               else if (XGetPixel (image,j,i) != myBgPixel)
                  data |= (1<<bit_count);

               if (++bit_count == 8)
               {
                  if (byte_count++ == 12)
                  {
                     byte_count = 1;
                     fprintf (fp, "\n   ");
                  }
                  fprintf (fp, "0x%c", hexValue[(data>>4) & 0xf]);

                  if (i == h-1 && j == w-1)
                     fprintf (fp, "%c};\n", hexValue[data & 0xf]);
                  else
                     fprintf (fp, "%c, ", hexValue[data & 0xf]);

                  bit_count = 0;
                  data = 0;
               }
            }
            if ((w % 8) != 0)
            {
               if (byte_count++ == 12)
               {
                  byte_count = 1;
                  fprintf (fp, "\n   ");
               }
               for (j = (w%8); j < 8; j++) data |= (1<<j);

               fprintf (fp, "0x%c", hexValue[(data>>4) & 0xf]);
               if (i == h-1)
                  fprintf (fp, "%c};\n", hexValue[data & 0xf]);
               else
                  fprintf (fp, "%c, ", hexValue[data & 0xf]);
            }
         }
      }
   }
   fclose (fp);

   XDestroyImage (image);
   XFreePixmap (mainDisplay, pixmap);

   if (short_name)
   {
      if (colorDump)
         sprintf (msg, "XPM file printed into '%s'.", rest);
      else
         sprintf (msg, "XBM file printed into '%s'.", rest);
   }
   else
   {
      if (colorDump)
         sprintf (msg, "XPM file printed into '%s'.", xbm_file_name);
      else
         sprintf (msg, "XBM file printed into '%s'.", xbm_file_name);
   }
   Msg (msg);
}

void DumpXBmObj (FP, ObjPtr)
   FILE			* FP;
   struct ObjRec	* ObjPtr;
{  /* called when printing in the PostScript format */
   register int		j, data, bit_count;
   int			i, ltx, lty, rbx, rby, w, h, block_w, block_h;
   int			num_nibbles, nibble_count, fill, h_blocks, v_blocks;
   int			row, col, x, y, nibbles_per_row, image_w, image_h;
   int			color_index, rotate, flip;
   int			orig_x, orig_y;
   Pixmap		bitmap;
   char			* xbm_data;
   XImage		* image=NULL;
   struct XBmRec	* xbm_ptr=ObjPtr->detail.xbm;
   struct MtrxRec	mtrx;

   if (xbm_ptr->real_type == XBM_EPS)
   {
      DumpEPSObj (FP, ObjPtr);
      return;
   }

   ltx = ObjPtr->obbox.ltx;
   lty = ObjPtr->obbox.lty;
   rbx = ObjPtr->obbox.rbx;
   rby = ObjPtr->obbox.rby;


   bitmap = xbm_ptr->bitmap;
   fill = xbm_ptr->fill;
   rotate = xbm_ptr->rotate;
   flip = xbm_ptr->flip;
   image_w = xbm_ptr->image_w;
   image_h = xbm_ptr->image_h;

   w = rbx - ltx;
   h = rby - lty;

   if (!PRTGIF)
   {
      if ((image = xbm_ptr->image) == NULL)
         if ((image = xbm_ptr->image = XGetImage (mainDisplay, bitmap, 0, 0,
               image_w, image_h, 1, ZPixmap)) == NULL)
         {
            Msg ("XGetImage() failed!  May have run out of memory!");
            Msg ("X bitmap object skipped for printing.");
            return;
         }
   }

   color_index = ObjPtr->color;

   mtrx.image_w = (float)image_w; mtrx.image_h = (float)image_h;
   mtrx.w = (float)w; mtrx.h = (float)h;
   mtrx.rotate = rotate; mtrx.flip = flip;

   CalcTransform (&mtrx);

   orig_x = (mtrx.transformed_w >= 0.0) ? ltx : ltx+w;
   orig_y = (mtrx.transformed_h >= 0.0) ? lty : lty+h;

   fprintf (FP, "%% XBM\n");
   if (colorDump)
   {
      fprintf (FP, "%.3f %.3f %.3f setrgbcolor\n",
            ((float)tgifColors[color_index].red/maxRGB),
            ((float)tgifColors[color_index].green/maxRGB),
            ((float)tgifColors[color_index].blue/maxRGB));
   }
   else
      fprintf (FP, "0 setgray\n");

   switch (fill)
   {
      case NONEPAT: break;
      case SOLIDPAT:
         fprintf (FP, "newpath\n");
         fprintf (FP, "   %1d %1d moveto ", ltx, lty);
         fprintf (FP, "%1d %1d lineto ", rbx, lty);
         fprintf (FP, "%1d %1d lineto ", rbx, rby);
         fprintf (FP, "%1d %1d lineto\n", ltx, rby);
         fprintf (FP, "closepath fill\n\n");
         break;
      case BACKPAT:
         fprintf (FP, "newpath\n");
         fprintf (FP, "   %1d %1d moveto ", ltx, lty);
         fprintf (FP, "%1d %1d lineto ", rbx, lty);
         fprintf (FP, "%1d %1d lineto ", rbx, rby);
         fprintf (FP, "%1d %1d lineto\n", ltx, rby);
         fprintf (FP, "closepath 1 setgray fill\n\n");
         if (colorDump)
            fprintf (FP, "%.3f %.3f %.3f setrgbcolor\n",
                  ((float)tgifColors[color_index].red/maxRGB),
                  ((float)tgifColors[color_index].green/maxRGB),
                  ((float)tgifColors[color_index].blue/maxRGB));
         else
            fprintf (FP, "0 setgray\n");
         break;
      default:
         /* patterned */
         fprintf (FP, "gsave\n");
         if (!colorDump)
         {
            GrayCheck (fill);
            if (useGray)
               fprintf (FP, "   %s setgray\n", GrayStr(fill));
            else
               fprintf (FP, "   pat%1d %s\n", fill, patternStr);
         }
         else
         {
            fprintf (FP, "   newpath\n");
            fprintf (FP, "      %1d %1d moveto ", ltx, lty);
            fprintf (FP, "%1d %1d lineto ", rbx, lty);
            fprintf (FP, "%1d %1d lineto ", rbx, rby);
            fprintf (FP, "%1d %1d lineto\n", ltx, rby);
            fprintf (FP, "   closepath 1 setgray fill\n\n");
            fprintf (FP, "   %.3f %.3f %.3f setrgbcolor\n",
                  ((float)tgifColors[color_index].red/maxRGB),
                  ((float)tgifColors[color_index].green/maxRGB),
                  ((float)tgifColors[color_index].blue/maxRGB));
         }
         fprintf (FP, "   newpath\n");
         fprintf (FP, "      %1d %1d moveto ", ltx, lty);
         fprintf (FP, "%1d %1d lineto ", rbx, lty);
         fprintf (FP, "%1d %1d lineto ", rbx, rby);
         fprintf (FP, "%1d %1d lineto\n", ltx, rby);
         if (colorDump)
         {
            fprintf (FP, "   closepath eoclip\n");
            DumpPatFill (FP, fill, 8, ObjPtr->bbox, "   ");
         }
         else
            fprintf (FP, "   closepath fill\n");
         fprintf (FP, "grestore\n\n");
         break;
   }

   h_blocks = ((image_w&0xff) == 0) ? (image_w>>8) : ((image_w>>8)+1);
   v_blocks = ((image_h&0xff) == 0) ? (image_h>>8) : ((image_h>>8)+1);
   nibbles_per_row = ((image_w%4)==0) ? (int)(image_w>>2) : (int)(image_w>>2)+1;

   fprintf (FP, "gsave\n");
   fprintf (FP, "   %1d %1d translate %.3f %.3f scale %1d rotate\n\n",
         orig_x, orig_y, mtrx.dump_h_scale, mtrx.dump_v_scale, mtrx.degree);

   for (row = 0; row < v_blocks; row++)
   {
      y = row<<8;
      block_h = (row == v_blocks-1) ? image_h-y : 0x100;

      for (col = 0; col < h_blocks; col++)
      {
         x = col<<8;
         block_w = (col == h_blocks-1) ? image_w-x : 0x100;

         num_nibbles = ((block_w%4) == 0) ? (int)(block_w>>2) :
               (int)(block_w>>2)+1;

         fprintf (FP, "   gsave\n");
         fprintf (FP, "   %1d %1d translate\n", x, y);
         fprintf (FP, "   %1d %1d true [1 0 0 1 0 0]\n   {<", block_w, block_h);
         if (PRTGIF)
         {
            xbm_data = ObjPtr->detail.xbm->data;

            nibble_count = 0;
            for (i = 0; i < block_h; i++)
            {
               for (j = 0; j < num_nibbles; j++)
               {
                  if (nibble_count++ == 64)
                  {
                     nibble_count = 1;
                     fprintf (FP, "\n     ");
                  }
                  fprintf (FP, "%c",
                        xbm_data[(i+y)*nibbles_per_row+j+(x>>2)]);
               }
               if ((num_nibbles & 0x1) == 1)
               {
                  if (nibble_count++ == 64)
                  {
                     nibble_count = 1;
                     fprintf (FP, "\n     ");
                  }
                  fprintf (FP, "0");
               }
            }
         }
         else
         {
            nibble_count = 0;
            for (i = 0; i < block_h; i++)
            {
               bit_count = 0;
               data = 0;
               for (j = 0; j < block_w; j++)
               {
                  data = (XGetPixel (image, j+x, i+y) == 1) ? (data<<1) | 1 :
                        (data<<1);
                  if (++bit_count == 4)
                  {
                     if (nibble_count++ == 64)
                     {
                        nibble_count = 1;
                        fprintf (FP, "\n     ");
                     }
                     fprintf (FP, "%c", hexValue[data]);
                     bit_count = 0;
                     data = 0;
                  }
               }
               if ((block_w % 4) != 0)
               {
                  data <<= (4 - (block_w % 4));
                  if (nibble_count++ == 64)
                  {
                     nibble_count = 1;
                     fprintf (FP, "\n     ");
                  }
                  fprintf (FP, "%c", hexValue[data]);
               }
               if ((num_nibbles & 0x1) == 1)
               {
                  if (nibble_count++ == 64)
                  {
                     nibble_count = 1;
                     fprintf (FP, "\n     ");
                  }
                  fprintf (FP, "0");
               }
            }
         }
         fprintf (FP, ">}\n");
         fprintf (FP, "   imagemask\n");
         fprintf (FP, "   grestore\n");
         if (row!=v_blocks-1 || col!=h_blocks-1) fprintf (FP, "\n");
      }
   }
   fprintf (FP, "grestore\n");
   fprintf (FP, "\n");
}

int NeedsToCacheXBmObj (ObjPtr)
   struct ObjRec	* ObjPtr;
{
   register struct XBmRec	* xbm_ptr = ObjPtr->detail.xbm;
   int				w, h;

   w = ObjPtr->obbox.rbx - ObjPtr->obbox.ltx;
   h = ObjPtr->obbox.rby - ObjPtr->obbox.lty;

   return (zoomScale != 0 ||
      xbm_ptr->image_w != w || xbm_ptr->image_h != h ||
      xbm_ptr->rotate != ROTATE0 || xbm_ptr->flip != 0
   );
}

static
void DrawHiddenXBm (win, x, y, w, h, s)
   Window	win;
   int		x, y, w, h;
   char		* s;
{
   int		str_w, len, sx, sy;
   XGCValues	values;

   values.foreground = myFgPixel;
   values.function = GXcopy;;
   values.fill_style = FillSolid;
#ifdef NO_THIN_LINE
   values.line_width = 1;
#else
   values.line_width = 0;
#endif
   values.font = rulerFontPtr->fid;
   values.line_style = LineSolid;

   XChangeGC (mainDisplay, drawGC,
         GCForeground | GCFunction | GCFillStyle | GCLineWidth | GCFont |
         GCLineStyle, &values);
   XDrawRectangle (mainDisplay, win, drawGC, x, y, w, h);

   len = strlen (s);
   str_w = rulerFontWidth*len;
   if (str_w < w && rulerFontHeight < h)
   {
      sx = x + ((w-str_w)>>1);
      sy = y + ((h-rulerFontHeight)>>1);
      XDrawString (mainDisplay, win, drawGC, sx, sy+rulerFontAsc, s, len);
   }
   XSetFont (mainDisplay, drawGC, canvasFontPtr->fid);
}

void DrawXBmObj (win, XOff, YOff, ObjPtr)
   Window		win;
   struct ObjRec	* ObjPtr;
{
   int			ltx, lty, rbx, rby, w, h, scr_w, scr_h;
   int			real_x_off, real_y_off;
   int			fill;
   char			s[80];
   struct XBmRec	* xbm_ptr = ObjPtr->detail.xbm;
   XGCValues		values;

   fill = xbm_ptr->fill;

   w = ObjPtr->obbox.rbx - ObjPtr->obbox.ltx;
   h = ObjPtr->obbox.rby - ObjPtr->obbox.lty;

   real_x_off = (zoomedIn ? XOff : (XOff>>zoomScale)<<zoomScale);
   real_y_off = (zoomedIn ? YOff : (YOff>>zoomScale)<<zoomScale);
   ltx = ZOOMED_SIZE(ObjPtr->obbox.ltx - real_x_off);
   lty = ZOOMED_SIZE(ObjPtr->obbox.lty - real_y_off);
   rbx = ZOOMED_SIZE(ObjPtr->obbox.rbx - real_x_off);
   rby = ZOOMED_SIZE(ObjPtr->obbox.rby - real_y_off);
   scr_w = rbx - ltx;
   scr_h = rby - lty;

   if (xbm_ptr->real_type==XBM_XBM && fill != NONEPAT)
   {
      values.foreground = (fill == 2) ? myBgPixel : colorPixels[ObjPtr->color];
      values.function = GXcopy;
      values.fill_style = FillOpaqueStippled;
      values.stipple = patPixmap[fill];
      XChangeGC (mainDisplay, drawGC,
            GCForeground | GCFunction | GCFillStyle | GCStipple, &values);
      XFillRectangle (mainDisplay, win, drawGC, ltx, lty, scr_w, scr_h);
   }

   if (!mapShown)
   {
      if (xbm_ptr->real_type == XBM_EPS)
         DrawHiddenXBm (win, ltx, lty, scr_w, scr_h, xbm_ptr->filename);
      else
      {
         sprintf (s, "(%1dx%1d)", xbm_ptr->image_w, xbm_ptr->image_h);
         DrawHiddenXBm (win, ltx, lty, scr_w, scr_h, s);
      }
      return;
   }

   if (NeedsToCacheXBmObj (ObjPtr)) MakeCachedBitmap (ObjPtr);

   if (xbm_ptr->real_type==XBM_XBM || xbm_ptr->real_type==XBM_EPS &&
         xbm_ptr->bitmap!=None)
   {
      values.foreground = colorPixels[ObjPtr->color];
      values.function = GXcopy;
      values.fill_style = FillStippled;
      values.ts_x_origin = ltx;
      values.ts_y_origin = lty;

      if (zoomScale==0 && xbm_ptr->rotate==ROTATE0 && xbm_ptr->flip==0 &&
            xbm_ptr->image_w==w && xbm_ptr->image_h==h)
         values.stipple = xbm_ptr->bitmap;
      else
      {
         if (xbm_ptr->cached_bitmap == None) return;
         values.stipple = xbm_ptr->cached_bitmap;
      }

      XChangeGC (mainDisplay, drawGC,
            GCForeground | GCFunction | GCFillStyle | GCStipple |
            GCTileStipXOrigin | GCTileStipYOrigin, &values);
      XFillRectangle (mainDisplay, win, drawGC, ltx, lty, scr_w, scr_h);
      XSetTSOrigin (mainDisplay, drawGC, 0, 0);
   }
   else if (xbm_ptr->real_type==XBM_EPS && xbm_ptr->bitmap==None)
      DrawHiddenXBm (win, ltx, lty, rbx-ltx, rby-lty, xbm_ptr->filename);
}

struct ObjRec * CreateXBmObj (ImageW, ImageH, W, H, bitmap, image)
   int		ImageW, ImageH, W, H;
   Pixmap	bitmap;
   XImage	* image;
{
   struct XBmRec	* xbm_ptr;
   struct ObjRec	* obj_ptr;

   xbm_ptr = (struct XBmRec *) calloc (1, sizeof(struct XBmRec));

   xbm_ptr->image = image;
   xbm_ptr->image_w = ImageW;
   xbm_ptr->image_h = ImageH;
   xbm_ptr->bitmap = bitmap;
   xbm_ptr->data = NULL;

   xbm_ptr->eps_w = xbm_ptr->eps_h = 0;

   xbm_ptr->fill = objFill;
   xbm_ptr->rotate = xbm_ptr->cached_rotate = ROTATE0;
   xbm_ptr->flip = xbm_ptr->cached_flip = 0;
   xbm_ptr->cached_zoom = 0;
   xbm_ptr->cached_bitmap = None;
   xbm_ptr->cached_w = xbm_ptr->cached_h = 0;

   xbm_ptr->real_type = XBM_XBM;
   xbm_ptr->filename = NULL;
   xbm_ptr->epsflines = NULL;
   xbm_ptr->num_epsf_lines = 0;

   obj_ptr = (struct ObjRec *) calloc (1, sizeof(struct ObjRec));

   obj_ptr->bbox.ltx = obj_ptr->obbox.ltx = obj_ptr->x = drawOrigX;
   obj_ptr->bbox.lty = obj_ptr->obbox.lty = obj_ptr->y = drawOrigY;
   obj_ptr->bbox.rbx = obj_ptr->obbox.rbx = W + drawOrigX;
   obj_ptr->bbox.rby = obj_ptr->obbox.rby = H + drawOrigY;
   obj_ptr->type = OBJ_XBM;
   obj_ptr->color = colorIndex;
   obj_ptr->id = objId++;
   obj_ptr->dirty = FALSE;
   obj_ptr->rotation = 0;
   obj_ptr->locked = FALSE;
   obj_ptr->detail.xbm = xbm_ptr;
   obj_ptr->fattr = obj_ptr->lattr = NULL;

   if (importXBmRV) InvertXBmObject (obj_ptr);
   return (obj_ptr);
}

static
void DumpXBmData (FP, bitmap, image, W, H)
   FILE		* FP;
   Pixmap	bitmap;
   XImage	* * image;
   int		W, H;
{
   register int	nibble_count, bit_count, data, i, j;

   if (*image == NULL)
      if ((*image = XGetImage (mainDisplay,bitmap,0,0,W,H,1,ZPixmap)) == NULL)
      {
         Msg ("XGetImage() failed!  May have run out of memory!");
         Msg ("Output file maybe corrupted.");
         return;
      }

   nibble_count = 0;

   for (i = 0; i < H; i++)
   {
      bit_count = 0;
      data = 0;

      for (j = 0; j < W; j++)
      {
         data = (XGetPixel (*image, j, i) == 1) ? (data<<1) | 1 : (data<<1);

         if (++bit_count == 4)
         {
            if (nibble_count++ == 64)
            {
               nibble_count = 1;
               fprintf (FP, "\n     ");
            }
            fprintf (FP, "%c", hexValue[data]);
            bit_count = 0;
            data = 0;
         }
      }
      if ((W % 4) != 0)
      {
         data <<= (4 - (W % 4));
         if (nibble_count++ == 64)
         {
            nibble_count = 1;
            fprintf (FP, "\n     ");
         }
         fprintf (FP, "%c", hexValue[data]);
      }
   }
}

void SaveXBmObj (FP, ObjPtr)
   FILE			* FP;
   struct ObjRec	* ObjPtr;
{
   int			ltx, lty, rbx, rby, image_w, image_h;
   int			no_bitmap;
   struct XBmRec	* xbm_ptr=ObjPtr->detail.xbm;

   no_bitmap = (xbm_ptr->real_type==XBM_EPS);

   ltx = ObjPtr->obbox.ltx; lty = ObjPtr->obbox.lty;
   rbx = ObjPtr->obbox.rbx; rby = ObjPtr->obbox.rby;
   if (no_bitmap)
   {
      image_w = xbm_ptr->eps_w;
      image_h = xbm_ptr->eps_h;
   }
   else
   {
      image_w = xbm_ptr->image_w;
      image_h = xbm_ptr->image_h;
   }
   fprintf (FP, "xbm('%s',", colorMenuItems[ObjPtr->color]);
   fprintf (FP, "%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d,%1d,\n    ",
         ltx, lty, rbx, rby, xbm_ptr->fill, ObjPtr->id, ObjPtr->rotation,
         image_w, image_h, xbm_ptr->rotate, xbm_ptr->flip, xbm_ptr->real_type,
         xbm_ptr->llx, xbm_ptr->lly, xbm_ptr->urx, xbm_ptr->ury,
         no_bitmap, ObjPtr->locked);
   switch (xbm_ptr->real_type)
   {
      case XBM_XBM: fprintf (FP, "\"\",\"\","); break;
      case XBM_EPS:
         fprintf (FP, "\"%s\",\"%s\",", xbm_ptr->write_date, xbm_ptr->filename);
         break;
   }
   if (!no_bitmap)
   {
      fprintf (FP, "\n    \"");
      DumpXBmData (FP, xbm_ptr->bitmap, &(xbm_ptr->image), image_w, image_h);
      fprintf (FP, "\",");
   }

   SaveAttrs (FP, ObjPtr->lattr);
   fprintf (FP, ")");
}

#define GETVALUE(val,name) ScanValue("%d", (char *) &(val), name, "xbitmap")

void ReadXBmObj (FP, Inbuf, ObjPtr)
   FILE			* FP;
   char			* Inbuf;
   struct ObjRec	* * ObjPtr;
{
   struct XBmRec	* xbm_ptr;
   char			color_str[20], * s, inbuf[MAXSTRING], * c_ptr;
   int			ltx, lty, rbx, rby, i, j, k, data=0, color_index;
   int			nibble_count, bit_count, num_nibbles, fill;
   int			rotation, new_alloc, id=0, image_w, image_h;
   int			rotate=ROTATE0, flip=0, real_type=XBM_XBM, len;
   int			no_bitmap=FALSE, llx=0, lly=0, urx=0, ury=0;
   int			locked=FALSE;
   Pixmap		bitmap;
   char			msg[MAXSTRING];
   char			* xbm_data, write_date[32], * filename;
   XImage		* image;

   *ObjPtr = NULL;

   s = FindChar ('(', Inbuf);
   s = ParseStr (s, ',', color_str);

   InitScan (s, "\t\n, ");

   rotation = 0;
   if (fileVersion <= 8)
   {
      sprintf (inbuf, "Invalid xbitmap version (%1d).", fileVersion);
      if (PRTGIF)
         fprintf (stderr, "%s\n", inbuf);
      else
         Msg (inbuf);
      return;
   }
   else if (fileVersion <= 13)
   {
      if (GETVALUE (ltx,      "ltx") == INVALID ||
          GETVALUE (lty,      "lty") == INVALID ||
          GETVALUE (rbx,      "rbx") == INVALID ||
          GETVALUE (rby,      "rby") == INVALID ||
          GETVALUE (fill,     "fill") == INVALID ||
          GETVALUE (id,       "id") == INVALID)
      {
         return;
      }
      if (id >= objId) objId = id+1;
   }
   else if (fileVersion <= 22)
   {
      if (GETVALUE (ltx,      "ltx") == INVALID ||
          GETVALUE (lty,      "lty") == INVALID ||
          GETVALUE (rbx,      "rbx") == INVALID ||
          GETVALUE (rby,      "rby") == INVALID ||
          GETVALUE (fill,     "fill") == INVALID ||
          GETVALUE (id,       "id") == INVALID ||
          GETVALUE (rotation, "rotation") == INVALID)
      {
         return;
      }
      if (id >= objId) objId = id+1;
   }
   else if (fileVersion <= 23)
   {
      if (GETVALUE (ltx,      "ltx") == INVALID ||
          GETVALUE (lty,      "lty") == INVALID ||
          GETVALUE (rbx,      "rbx") == INVALID ||
          GETVALUE (rby,      "rby") == INVALID ||
          GETVALUE (fill,     "fill") == INVALID ||
          GETVALUE (id,       "id") == INVALID ||
          GETVALUE (rotation, "rotation") == INVALID ||
          GETVALUE (image_w,  "image_w") == INVALID ||
          GETVALUE (image_h,  "image_h") == INVALID ||
          GETVALUE (rotate,   "rotate") == INVALID ||
          GETVALUE (flip,     "flip") == INVALID)
      {
         return;
      }
      if (id >= objId) objId = id+1;
   }
   else if (fileVersion <= 25)
   {
      if (GETVALUE (ltx,       "ltx") == INVALID ||
          GETVALUE (lty,       "lty") == INVALID ||
          GETVALUE (rbx,       "rbx") == INVALID ||
          GETVALUE (rby,       "rby") == INVALID ||
          GETVALUE (fill,      "fill") == INVALID ||
          GETVALUE (id,        "id") == INVALID ||
          GETVALUE (rotation,  "rotation") == INVALID ||
          GETVALUE (image_w,   "image_w") == INVALID ||
          GETVALUE (image_h,   "image_h") == INVALID ||
          GETVALUE (rotate,    "rotate") == INVALID ||
          GETVALUE (flip,      "flip") == INVALID ||
          GETVALUE (real_type, "real_type") == INVALID ||
          GETVALUE (llx,       "llx") == INVALID ||
          GETVALUE (lly,       "lly") == INVALID ||
          GETVALUE (urx,       "urx") == INVALID ||
          GETVALUE (ury,       "ury") == INVALID ||
          GETVALUE (no_bitmap, "no_bitmap") == INVALID)
      {
         return;
      }
      if (id >= objId) objId = id+1;
   }
   else
   {
      if (GETVALUE (ltx,       "ltx") == INVALID ||
          GETVALUE (lty,       "lty") == INVALID ||
          GETVALUE (rbx,       "rbx") == INVALID ||
          GETVALUE (rby,       "rby") == INVALID ||
          GETVALUE (fill,      "fill") == INVALID ||
          GETVALUE (id,        "id") == INVALID ||
          GETVALUE (rotation,  "rotation") == INVALID ||
          GETVALUE (image_w,   "image_w") == INVALID ||
          GETVALUE (image_h,   "image_h") == INVALID ||
          GETVALUE (rotate,    "rotate") == INVALID ||
          GETVALUE (flip,      "flip") == INVALID ||
          GETVALUE (real_type, "real_type") == INVALID ||
          GETVALUE (llx,       "llx") == INVALID ||
          GETVALUE (lly,       "lly") == INVALID ||
          GETVALUE (urx,       "urx") == INVALID ||
          GETVALUE (ury,       "ury") == INVALID ||
          GETVALUE (no_bitmap, "no_bitmap") == INVALID ||
          GETVALUE (locked,    "locked") == INVALID)
      {
         return;
      }
      if (id >= objId) objId = id+1;
   }
   if (fileVersion <= 22)
   {
      image_w = rbx-ltx;
      image_h = rby-lty;
      rotate=ROTATE0;
      flip=0;
   }
   if (fileVersion >= 24)
   {
      char	* tmp_str, * s, * s1;

      fgets (inbuf, MAXSTRING, FP);
      scanLineNum++;

      tmp_str = FindChar ('"', inbuf);
      s = s1 = ReadString (tmp_str);
      s1++;
      *(--s) = '\0';
      len = strlen (tmp_str);
      if (len > 24)
      {
         (void) sprintf (msg, "%s, %d:  %s too long in %s",
               scanFileName, scanLineNum, "write_date", "xbm");
         if (PRTGIF)
            fprintf (stderr, "%s\n", msg);
         else
            Msg (msg);
      }
      strcpy (write_date, tmp_str);

      tmp_str = FindChar ('"', s1);
      s = ReadString (tmp_str);
      *(--s) = '\0';
      len = strlen (tmp_str);
      filename = (char *) calloc (len+1, sizeof(char));
      strcpy (filename, tmp_str);
   }
   else
   {
      *write_date = '\0';
      filename = NULL;
   }
   fill = UpgradePenFill (fill);

   *ObjPtr = (struct ObjRec *) calloc (1, sizeof(struct ObjRec));
   xbm_ptr = (struct XBmRec *) calloc (1, sizeof(struct XBmRec));

   color_index = QuickFindColorIndex (color_str, &new_alloc);

   (*ObjPtr)->color = color_index;
   (*ObjPtr)->dirty = FALSE;
   (*ObjPtr)->id = id;
   (*ObjPtr)->rotation = rotation;
   (*ObjPtr)->locked = locked;
   (*ObjPtr)->type = OBJ_XBM;
   (*ObjPtr)->obbox.ltx = (*ObjPtr)->bbox.ltx = (*ObjPtr)->x = ltx;
   (*ObjPtr)->obbox.lty = (*ObjPtr)->bbox.lty = (*ObjPtr)->y = lty;
   (*ObjPtr)->obbox.rbx = (*ObjPtr)->bbox.rbx = rbx;
   (*ObjPtr)->obbox.rby = (*ObjPtr)->bbox.rby = rby;
   (*ObjPtr)->detail.xbm = xbm_ptr;

   xbm_ptr->bitmap = NULL;
   xbm_ptr->image = NULL;
   xbm_ptr->cached_bitmap = None;
   xbm_ptr->cached_zoom = 0;
   xbm_ptr->data = NULL;
   xbm_ptr->fill = fill;
   xbm_ptr->rotate = rotate;
   xbm_ptr->flip = flip;
   xbm_ptr->cached_rotate = INVALID;
   xbm_ptr->cached_flip = 0;
   xbm_ptr->cached_w = xbm_ptr->cached_h = 0;
   xbm_ptr->image_w = image_w;
   xbm_ptr->image_h = image_h;
   xbm_ptr->llx = llx; xbm_ptr->lly = lly;
   xbm_ptr->urx = urx; xbm_ptr->ury = ury;

   xbm_ptr->real_type = real_type;
   strcpy (xbm_ptr->write_date, write_date);
   xbm_ptr->filename = filename;
   xbm_ptr->epsflines = NULL;
   xbm_ptr->num_epsf_lines = 0;

   if (no_bitmap && xbm_ptr->real_type==XBM_EPS)
   {
      float	file_llx=0.0, file_lly=0.0, file_urx=0.0, file_ury=0.0;

      xbm_ptr->eps_w = xbm_ptr->image_w;
      xbm_ptr->eps_h = xbm_ptr->image_h;
      xbm_ptr->image_w = 0;
      xbm_ptr->image_h = 0;

      if (PRTGIF) return;

      if (MyReadEPSFile (xbm_ptr->filename, &image_w,  &image_h,
            &(xbm_ptr->bitmap), &(xbm_ptr->image), &(xbm_ptr->num_epsf_lines),
            &(xbm_ptr->epsflines), &(xbm_ptr->epsf_level), &file_llx, &file_lly,
            &file_urx, &file_ury, xbm_ptr->write_date) != BitmapSuccess)
         return;
      if (xbm_ptr->bitmap == None)
      {
         xbm_ptr->eps_w = (file_urx >= file_llx) ?
               (int)(file_urx-file_llx) : (int)(file_llx-file_urx);
         xbm_ptr->eps_h = (file_ury >= file_lly) ?
               (int)(file_ury-file_lly) : (int)(file_lly-file_ury);
      }
      else
      {
         xbm_ptr->eps_w = xbm_ptr->image_w = image_w;
         xbm_ptr->eps_h = xbm_ptr->image_h = image_h;
      }
      xbm_ptr->llx = (int)(file_llx*1000.0);
      xbm_ptr->lly = (int)(file_lly*1000.0);
      xbm_ptr->urx = (int)(file_urx*1000.0);
      xbm_ptr->ury = (int)(file_ury*1000.0);
      return;
   }
   if (xbm_ptr->real_type == XBM_EPS)
   {
      xbm_ptr->eps_w = xbm_ptr->image_w;
      xbm_ptr->eps_h = xbm_ptr->image_h;
   }
   else
      xbm_ptr->eps_w = xbm_ptr->eps_h = 0;

   num_nibbles = ((image_w % 4) == 0) ? (int)(image_w>>2) : (int)(image_w>>2)+1;

   fgets (inbuf, MAXSTRING, FP);
   scanLineNum++;
   c_ptr = &inbuf[5];
   nibble_count = 0;

   if (PRTGIF)
   {
      xbm_data = (char *) calloc (image_h*num_nibbles, sizeof(char));
      for (i = 0; i < image_h; i++)
      {
         for (j = 0; j < num_nibbles; j++)
         {
            if (nibble_count++ == 64)
            {
               fgets (inbuf, MAXSTRING, FP);
               scanLineNum++;
               c_ptr = &inbuf[5];
               nibble_count = 1;
            }
            xbm_data[i*num_nibbles+j] = *c_ptr++;
         }
      }
      xbm_ptr->data = xbm_data;
   }
   else
   {
      bitmap = XCreatePixmap (mainDisplay, dummyBitmap, image_w, image_h, 1);
      XFillRectangle (mainDisplay, bitmap, xbmGC, 0, 0, image_w, image_h);
      image = XGetImage (mainDisplay,bitmap,0,0,image_w,image_h,1,ZPixmap);

      for (i = 0; i < image_h; i++)
      {
         bit_count = 0;
         for (j = 0; j < num_nibbles; j++)
         {
            if (nibble_count++ == 64)
            {
               fgets (inbuf, MAXSTRING, FP);
               scanLineNum++;
               c_ptr = &inbuf[5];
               nibble_count = 1;
            }

            if (*c_ptr >= '0' && *c_ptr <= '9')
               data = (int)(*c_ptr++) - (int)('0');
            else if (*c_ptr >= 'a' && *c_ptr <= 'f')
               data = (int)(*c_ptr++) - (int)('a') + 10;

            for (k = 0; k < 4; k++)
            {
               if (bit_count++ == image_w) break;

               if (data & (1<<(3-k)))
                  XPutPixel (image, j*4+k, i, 1);
            }
         }
      }
      XPutImage (mainDisplay,bitmap,xbmGC,image,0,0,0,0,image_w,image_h);
      xbm_ptr->bitmap = bitmap;
      xbm_ptr->image = image;
   }
}

void FreeXBmObj (ObjPtr)
   struct ObjRec	* ObjPtr;
{
   register struct XBmRec	* xbm_ptr = ObjPtr->detail.xbm;
   register int			i;

   if (xbm_ptr->bitmap != NULL) XFreePixmap (mainDisplay, xbm_ptr->bitmap);
   if (xbm_ptr->image != NULL) XDestroyImage (xbm_ptr->image);
   if (xbm_ptr->cached_bitmap != None)
      XFreePixmap (mainDisplay, xbm_ptr->cached_bitmap);
   xbm_ptr->bitmap = None;
   xbm_ptr->image = NULL;
   xbm_ptr->cached_bitmap = None;
   xbm_ptr->cached_zoom = 0;
   if (xbm_ptr->data != NULL) free (xbm_ptr->data);
   if (xbm_ptr->real_type == XBM_EPS)
   {
      if (xbm_ptr->filename != NULL) free (xbm_ptr->filename);
      for (i = 0; i < xbm_ptr->num_epsf_lines; i++)
         if (xbm_ptr->epsflines[i] != NULL)
            free (xbm_ptr->epsflines[i]);
      if (xbm_ptr->epsflines != NULL) free (xbm_ptr->epsflines);
   }
   free (xbm_ptr);
   free (ObjPtr);
}
